/*****************************************************************************/

/*
 *      winsock.c  -- directly attach a native WinSock interface.
 *
 * Developed for TNOS (brian@lantz.com)
 *
 */

static char rcsid[] OPTIONAL = "$Id: winsock.c,v 1.1.1.1 2004/02/09 21:53:56 blantz Exp $";

#ifdef __CYGWIN__

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <net/if.h>
#include "system.h"

#include <cygwin/in.h>

#undef LOCK
#define	_SOCKET_H
#define _SOCKADDR_H
#include "global.h"

#if defined(WINSOCK)

#include "iface.h"

#include "commands.h"
#include "devparam.h"
#include "trace.h"
#include "hardware.h"
#include "enet.h"
#include "pktdrvr.h"

typedef void * HANDLE;
#define _WINSOCK2_H
#include "pcap.h"


#define WINSOCK_MAX 10


/* winsock interface control block */
struct winsockif {
    struct iface *iface;

    int class;		/* device class (for dump) */

    long rxpkts;            /* receive packets */
    long txpkts;            /* transmit packets */
    long rxchar;            /* Received characters */
    long txchar;            /* Transmitted characters */
    
    struct proc *rx_raw;
    const u_char *packet;
    struct pcap_pkthdr hdr;
    pcap_t *pcap;
    char *devname;
};

static struct winsockif WinsockIf[WINSOCK_MAX];

static void winsock_rx_raw(int dev, void *v1, void *v2);


/*---------------------------------------------------------------------------*/

static int winsock_up(struct winsockif *wi)
{
char *cp;
char errbuf[PCAP_ERRBUF_SIZE];

    wi->pcap = pcap_open_live (wi->devname, BUFSIZ, 1, 5, errbuf);
    if (wi->pcap == NULL) {
        tprintf ("Error opening device: %s\n\t%s\n", wi->devname, errbuf);
        return -1;
    }

    cp = if_name (wi->iface, " rx raw");
    wi->rx_raw = newproc (cp, 768, winsock_rx_raw, 0, wi, NULL, 0);
    return 0;
}

/*---------------------------------------------------------------------------*/

static int winsock_down(struct winsockif *wi)
{
    if(wi->iface == NULL)
        return 0;

    killproc (wi->rx_raw);
    wi->rx_raw = NULLPROC;
    pcap_close (wi->pcap);
    return 0;
}

/*---------------------------------------------------------------------------*/

/* Asynchronous line I/O control */

static int32 winsock_ioctl(struct iface *ifp OPTIONAL, int cmd, int set OPTIONAL, int32 val OPTIONAL)
{
struct winsockif *wi;
    
    wi = WinsockIf + ifp->dev;
    switch(cmd){
        case PARAM_DOWN:        return winsock_down(wi) ? 0 : 1;
	case PARAM_UP:		return winsock_up(wi) ? 0 : 1;
        default:	        return -1;
	}
}

/*---------------------------------------------------------------------------*/

/*
 * Send routine for packet driver
 */

static int winsock_raw(struct iface *iface, struct mbuf *bp)
{
uint8 buf[1600]; /* should be enough */
uint8 *bufp = buf;
struct mbuf *bp2;
int cnt = 0;
int retval = 0;
struct winsockif *wi;
    
    wi = WinsockIf + iface->dev;

    dump (iface, IF_TRACE_OUT, wi->iface->type, bp);

    iface->rawsndcnt++;
    iface->lastsent = secclock();
    if (iface->trace & IF_TRACE_RAW)
	raw_dump(iface, -1, bp);

    if (wi->iface == NULL)
        free_p (bp);
    else {
	for (bp2 = bp; bp2; bp2 = bp2->next) {
            cnt += bp2->cnt;
            if (cnt > (int) sizeof(buf)) {
		(void) free_mbuf (bp);
		return -1;
            }
            memcpy (bufp, bp2->data, bp2->cnt);
            bufp += bp2->cnt;
	}
        retval = pcap_sendpacket (wi->pcap, buf, cnt);
            
	if (retval == 0) {
            wi->txpkts++;
            wi->txchar += cnt;
	}
        (void) free_mbuf(bp);
    }
    return retval;
}

/*---------------------------------------------------------------------------*/

static int winsock_rx_handler (struct winsockif *wi)
{
int i;
struct mbuf *bp;
struct phdr phdr;
    
    wi->packet = pcap_next (wi->pcap, &wi->hdr);
    if (wi->packet == NULL)
        return 0;
    
    i = wi->hdr.len;
    
    if ((bp = alloc_mbuf(i + sizeof (phdr) + 16)) == 0)        
        return -1;
    phdr.iface = wi->iface;
    phdr.type = wi->iface->type;
    memcpy (bp->data, &phdr, sizeof (phdr));
    
    memcpy (&bp->data[sizeof (phdr)], wi->packet, i);
    bp->cnt = (int16) (i + sizeof (phdr));
    wi->rxpkts++;
    wi->rxchar += i;
    if (wi->iface->trace & IF_TRACE_RAW)
	raw_dump(wi->iface, IF_TRACE_IN, bp);

    (void) enqueue (&Hopper, bp);
    return 1;
}


static void winsock_rx_raw (int i OPTIONAL, void *v1, void *v2 OPTIONAL)
{
    struct winsockif *wi = (struct winsockif *)v1;
    int retval = 1;
    server_disconnect_io ();

    for ( ; ; )	{
        kpause (__FILE__, __LINE__, 100);
        
        do {
            retval = winsock_rx_handler (wi);
        } while (retval == 1);
        
        if (retval == -1)
            break;
    }
    wi->rx_raw = NULLPROC;
}

/*---------------------------------------------------------------------------*/

static void winsock_status(struct iface *iface)
{
struct winsockif *wi;
    
    wi = WinsockIf + iface->dev;
    if (wi->iface != iface)
        return;

    tprintf("           Received:    Packets %8ld Chars %8ld\n"
            "           Transmitted: Packets %8ld Chars %8ld\n",
            wi->rxpkts, wi->rxchar, wi->txpkts, wi->txchar);
    tprintf("           Device:      %s\n", wi->devname);
}

/*---------------------------------------------------------------------------*/

static int winsock_detach(struct iface *ifp)
{
struct winsockif *wi;
    
    wi = WinsockIf + ifp->dev;
    if (ifp == NULL || wi->iface == NULL)
	return -1;      /* Not allocated */
    (void) winsock_down(wi);
    wi->iface = NULL;
    free (wi->devname);
    wi->devname = NULLCHAR;
    return 0;
}

/*---------------------------------------------------------------------------*/

int winsock_attach(int argc OPTIONAL, char *argv[], void *p OPTIONAL)
{
struct iface *ifp;
int dev;
struct winsockif *wi;

    if (if_lookup(argv[1]) != NULL) {
        tprintf("Interface %s already exists\n",argv[1]);
        return -1;
    }
    
    /* Find unused krnlif control block */
    for (dev=0; (dev < WINSOCK_MAX) && (WinsockIf[dev].iface); dev++)
		;
    if (dev >= WINSOCK_MAX) {
        tprintf ("Too many winsock interfaces\n");
	return -1;
    }
    wi = WinsockIf+dev;
    
    /* Create interface structure and fill in details */
    ifp = (struct iface *)callocw(1,sizeof(struct iface));
    ifp->addr = Ip_addr;
    ifp->name = strdup(argv[1]);
    ifp->mtu = MAXINT16;
    ifp->stop = winsock_detach;
    /* set encapsulation */
    (void) setencap (ifp, "Ethernet");
    ifp->ioctl = winsock_ioctl;
    ifp->raw = winsock_raw;
    ifp->show = winsock_status;
    ifp->hwaddr = mallocw(EADDR_LEN);
    gether (ifp->hwaddr, "0:2:3:4:5:6");
    ifp->dev = dev;
    ifp->xdev = dev;
	
    wi->iface = ifp;
    wi->devname = strdup (argv[2]);

    /* Link in the interface */
    ifp->next = Ifaces;
    Ifaces = ifp;

    return winsock_up(wi);
}

/*---------------------------------------------------------------------------*/


#endif /* WINSOCK */
#endif /* __CYGWIN__ */
